from typing import Iterable, Union, Dict, Tuple, Optional, NamedTuple, List
from enum import Enum, auto
from string import ascii_letters, digits

from gi.repository.GLib import Variant


INTERFACE = "org.mpris.MediaPlayer2"
NAME = "mprisServer"
MIME_TYPES = ["audio/mpeg", "application/ogg", "video/mpeg"]
BUS_TYPE = "session"
URI = ["file"]
DEFAULT_DESKTOP = ''

# typically, these are the props that dbus needs to be notified about
# upon specific state-change events.
ON_ENDED_PROPS = ['PlaybackStatus']
ON_VOLUME_PROPS = ['Volume', 'Metadata']
ON_PLAYBACK_PROPS = ['PlaybackStatus', 'Metadata']
ON_PLAYPAUSE_PROPS = ['PlaybackStatus']
ON_TITLE_PROPS = ['Metadata']
ON_OPTION_PROPS = ['LoopStatus', 'Shuffle', 'CanGoPrevious', 'CanGoNext']
ON_SEEK_PROPS = ['Position']

ON_TRACKS_PROPS = ['Tracks']
ON_PLAYLIST_PROPS = ['PlaylistCount', 'Orderings', 'ActivePlaylist']

DEFAULT_RATE = 1.0
PAUSE_RATE = 0
MIN_RATE = 1.0
MAX_RATE = 1.0

MUTE_VOL = 0
MAX_VOL = 1
BEGINNING = 0

DEFAULT_TRACK_ID = '/default/1'
DEFAULT_PLAYLIST_COUNT = 1
DEFAULT_ORDERINGS = ["Alphabetical", "User"]

# valid characters for a DBus name 
VALID_PUNC = '_'
VALID_CHARS = set(digits + ascii_letters + VALID_PUNC)
NAME_PREFIX = "Mpris_Server_"
RAND_CHARS = 5

DEFAULT_METADATA = {}

# type aliases
Microseconds = int
VolumeDecimal = float
RateDecimal = float
DbusTypes = Union[str, float, int, bool, list]
Metadata = Dict[str, DbusTypes]
DbusMetadata = Dict[str, Variant]
DbusObj = str
PlaylistEntry = Tuple[str, str, str]
PlaylistValidity = bool


# See https://docs.python.org/3/library/enum.html#using-automatic-values
class AutoName(Enum):
    def _generate_next_value_(name: str, *args, **kwargs) -> str:
        return name


class PlayState(AutoName):
    PLAYING = auto()
    PAUSED = auto()
    STOPPED = auto()


class Artist(NamedTuple):
    name: str = "Default Artist"


class Album(NamedTuple):
    name: str = "Default Album"
    artists: Tuple[Artist] = tuple()
    art_url: Optional[str] = None


class Track(NamedTuple):
    track_id: DbusObj = DEFAULT_TRACK_ID
    name: str = "Default Track"
    track_no: Optional[int] = None
    length: Microseconds = 0
    uri: Optional[str] = None
    artists: Tuple[Artist] = tuple()
    album: Optional[Album] = None
    art_url: Optional[str] = None
    disc_no: Optional[int] = None
    type: Optional[Enum] = None


def dbus_emit_changes(interface: 'MprisInterface',
                      changes: Iterable[str]):
    attr_vals = {attr: getattr(interface, attr)
                 for attr in changes}

    interface.PropertiesChanged(interface.INTERFACE, attr_vals, [])



from typing import List, Tuple
from abc import ABC

from .base import URI, MIME_TYPES, PlayState, DEFAULT_RATE, Microseconds, \
  VolumeDecimal, RateDecimal, Metadata, DbusObj, PlaylistEntry, \
  PlaylistValidity, DEFAULT_PLAYLIST_COUNT, DEFAULT_ORDERINGS, \
  DEFAULT_DESKTOP, Track


class RootAdapter(ABC):
  def can_quit(self) -> bool:
    pass

  def can_raise(self) -> bool:
    pass

  def can_fullscreen(self) -> bool:
    pass

  def has_tracklist(self) -> bool:
    pass

  def get_uri_schemes(self) -> List[str]:
    return URI

  def get_mime_types(self) -> List[str]:
    return MIME_TYPES

  def set_raise(self, val: bool):
    pass

  def quit(self):
    pass

  def get_fullscreen(self) -> bool:
    return False

  def set_fullscreen(self, val: bool):
    pass

  def get_desktop_entry(self) -> str:
    return DEFAULT_DESKTOP


class PlayerAdapter(ABC):
  def metadata(self) -> Metadata:
    """
    Implement this function to supply your own MPRIS Metadata.

    If this function is implemented, metadata won't be built from get_current_track().

    See: https://www.freedesktop.org/wiki/Specifications/mpris-spec/metadata/
    :return:
    """
    pass

  def get_current_track(self) -> Track:
    """
    This function is an artifact of forking the base MPRIS library to a generic interface.
    The base library expected Track-like objects to build metadata.

    If metadata() is implemented, this function won't be used to build MPRIS metadata.
    :return:
    """
    pass

  def get_current_position(self) -> Microseconds:
    pass

  def next(self):
    pass

  def previous(self):
    pass

  def pause(self):
    pass

  def resume(self):
    pass

  def stop(self):
    pass

  def play(self):
    pass

  def get_playstate(self) -> PlayState:
    pass

  def seek(self, time: Microseconds):
    pass

  def open_uri(self, uri: str):
    pass

  def is_repeating(self) -> bool:
    pass

  def is_playlist(self) -> bool:
    pass

  def set_repeating(self, val: bool):
    pass

  def set_loop_status(self, val: str):
    pass

  def get_rate(self) -> RateDecimal:
    return DEFAULT_RATE

  def set_rate(self, val: RateDecimal):
    pass

  def get_shuffle(self) -> bool:
    pass

  def set_shuffle(self, val: bool):
    pass

  def get_art_url(self, track: int) -> str:
    pass

  def get_volume(self) -> VolumeDecimal:
    pass

  def set_volume(self, val: VolumeDecimal):
    pass

  def is_mute(self) -> bool:
    pass

  def set_mute(self, val: bool):
    pass

  def can_go_next(self) -> bool:
    pass

  def can_go_previous(self) -> bool:
    pass

  def can_play(self) -> bool:
    pass

  def can_pause(self) -> bool:
    pass

  def can_seek(self) -> bool:
    pass

  def can_control(self) -> bool:
    pass

  def get_stream_title(self) -> str:
    pass

  def get_previous_track(self) -> Track:
    pass

  def get_next_track(self) -> Track:
    pass


class PlaylistAdapter(ABC):
  def activate_playlist(self, id: DbusObj):
    pass

  def get_playlists(self, index: int, max_count: int, order: str, reverse: bool) -> List[PlaylistEntry]:
    pass

  def get_playlist_count(self) -> int:
    return DEFAULT_PLAYLIST_COUNT

  def get_orderings(self) -> List[str]:
    return DEFAULT_ORDERINGS

  def get_active_playlist(self) -> Tuple[PlaylistValidity, PlaylistEntry]:
    pass


class TrackListAdapter(ABC):
  def get_tracks_metadata(self, track_ids: List[DbusObj]) -> Metadata:
    pass

  def add_track(self, uri: str, after_track: DbusObj, set_as_current: bool):
    pass

  def remove_track(self, track_id: DbusObj):
    pass

  def go_to(self, track_id: DbusObj):
    pass

  def get_tracks(self) -> List[DbusObj]:
    pass

  def can_edit_tracks(self) -> bool:
    pass


class MprisAdapter(RootAdapter, PlayerAdapter, PlaylistAdapter, TrackListAdapter, ABC):
  """
  MRPRIS interface for your application.

  The MPRIS implementation is supplied with information
  returned from this adapter.
  """

  def __init__(self, name: str = 'MprisAdapter'):
    self.name = name



from abc import ABC
from typing import List, Optional

from .base import dbus_emit_changes, ON_ENDED_PROPS, \
  ON_VOLUME_PROPS, ON_PLAYBACK_PROPS, ON_PLAYPAUSE_PROPS, \
  ON_TITLE_PROPS, Microseconds, ON_SEEK_PROPS, ON_OPTION_PROPS, \
  DbusObj, ON_PLAYLIST_PROPS, ON_TRACKS_PROPS, Metadata
from .interface import MprisInterface
from .player import Player
from .playlists import Playlists
from .root import Root
from .tracklist import TrackList


class BaseEventAdapter(ABC):
  def __init__(
    self,
    player: Player,
    root: Root,
    playlist: Optional[Playlists] = None,
    tracklist: Optional[TrackList] = None
  ):
    self.root = root
    self.player = player
    self.playlist = playlist
    self.tracklist = tracklist

  def emit_changes(self, interface: MprisInterface, changes: List[str]):
    dbus_emit_changes(interface, changes)


class PlayerEventAdapter(BaseEventAdapter, ABC):
  def emit_player_changes(self, changes: List[str]):
    self.emit_changes(self.player, changes)

  def on_ended(self):
    self.emit_player_changes(ON_ENDED_PROPS)

  def on_volume(self):
    self.emit_player_changes(ON_VOLUME_PROPS)

  def on_playback(self):
    self.emit_player_changes(ON_PLAYBACK_PROPS)

  def on_playpause(self):
    self.emit_player_changes(ON_PLAYPAUSE_PROPS)

  def on_title(self):
    self.emit_player_changes(ON_TITLE_PROPS)

  def on_seek(self, position: Microseconds):
    self.player.Seeked(position)
    self.emit_player_changes(ON_SEEK_PROPS)

  def on_options(self):
    self.emit_player_changes(ON_OPTION_PROPS)


class PlaylistsEventAdapter(BaseEventAdapter, ABC):
  def emit_playlist_changes(self, changes: List[str]):
    self.emit_changes(self.playlist, changes)

  def on_playlist_change(self, playlist_id: DbusObj):
    self.playlist.PlaylistChanged(playlist_id)
    self.emit_playlist_changes(ON_PLAYLIST_PROPS)


class TracklistEventAdapter(BaseEventAdapter, ABC):
  def emit_tracklist_changes(self, changes: List[str]):
    self.emit_changes(self.tracklist, changes)

  def on_list_replaced(self, tracks: List[DbusObj], current_track: DbusObj):
    self.tracklist.TrackListReplaced(tracks, current_track)
    self.emit_tracklist_changes(ON_TRACKS_PROPS)

  def on_track_added(self, metadata: Metadata, after_track: DbusObj):
    self.tracklist.TrackAdded(metadata, after_track)
    self.emit_tracklist_changes(ON_TRACKS_PROPS)

  def on_track_removed(self, track_id: DbusObj):
    self.tracklist.TrackRemoved(track_id)
    self.emit_tracklist_changes(ON_TRACKS_PROPS)

  def on_track_metadata_change(self, track_id: DbusObj, metadata: Metadata):
    self.tracklist.TrackMetadataChanged(track_id, metadata)
    self.emit_tracklist_changes(ON_TRACKS_PROPS)


class EventAdapter(PlayerEventAdapter, TracklistEventAdapter, PlaylistsEventAdapter, ABC):
  """
  Notify DBUS of state-change events in the media player.

  Implement this class and integrate it in your application to emit
  DBUS signals when there are state changes in the media player.
  """
  pass



import logging
from typing import List

from pydbus.generic import signal

from .base import Metadata, DbusObj
from .interface import MprisInterface

logger = logging.getLogger(__name__)


class TrackList(MprisInterface):
  """
  <node>
    <interface name="org.mpris.MediaPlayer2.TrackList">
      <method name="GetTracksMetadata">
        <arg name="TrackIds" type="ao" direction="in"/>
        <arg name="Metadata" type="aa{sv}" direction="out"/>
      </method>
      <method name="AddTrack">
        <arg name="Uri" type="s" direction="in"/>
        <arg name="AfterTrack" type="o" direction="in"/>
        <arg name="SetAsCurrent" type="b" direction="in"/>
      </method>
      <method name="RemoveTrack">
        <arg name="TrackId" type="o" direction="in"/>
      </method>
      <method name="GoTo">
        <arg name="TrackId" type="o" direction="in"/>
      </method>
      <signal name="TrackListReplaced">
        <arg name="Tracks" type="ao"/>
        <arg name="CurrentTrack" type="o"/>
      </signal>
      <signal name="TrackAdded">
        <arg name="Metadata" type="a{sv}"/>
        <arg name="AfterTrack" type="o"/>
      </signal>
      <signal name="TrackRemoved">
        <arg name="TrackId" type="o"/>
      </signal>
      <signal name="TrackMetadataChanged">
        <arg name="TrackId" type="o"/>
        <arg name="Metadata" type="a{sv}"/>
      </signal>
      <property name="Tracks" type="ao" access="read"/>
      <property name="CanEditTracks" type="b" access="read"/>
    </interface>
  </node>
  """

  INTERFACE = "org.mpris.MediaPlayer2.TrackList"
  TrackListReplaced = signal()
  TrackAdded = signal()
  TrackRemoved = signal()
  TrackMetadataChanged = signal()

  def GetTracksMetadata(self, track_ids: List[DbusObj]) -> Metadata:
    return self.adapter.get_tracks_metadata(track_ids)

  def AddTrack(self,
               uri: str,
               after_track: DbusObj,
               set_as_current: bool):
    self.adapter.add_track(uri, after_track, set_as_current)

  def RemoveTrack(self, track_id: DbusObj):
    self.adapter.remove_track(track_id)

  def GoTo(self, track_id: DbusObj):
    self.adapter.go_to(track_id)

  @property
  def Tracks(self) -> List[DbusObj]:
    return self.adapter.get_tracks()

  @property
  def CanEditTracks(self) -> bool:
    return self.adapter.can_edit_tracks()



from typing import Optional
import logging

from gi.repository.GLib import Variant
from pydbus.generic import signal

from .base import PlayState, MUTE_VOL, MAX_VOL, PAUSE_RATE, BEGINNING, \
    Microseconds, RateDecimal, VolumeDecimal, MAX_RATE, MIN_RATE, \
    Track, DbusMetadata, DEFAULT_METADATA
from .compat import get_dbus_metadata
from .interface import MprisInterface

logger = logging.getLogger(__name__)


class Player(MprisInterface):
    """
    <node>
      <interface name="org.mpris.MediaPlayer2.Player">
        <method name="Next"/>
        <method name="Previous"/>
        <method name="Pause"/>
        <method name="PlayPause"/>
        <method name="Stop"/>
        <method name="Play"/>
        <method name="Seek">
          <arg name="Offset" type="x" direction="in"/>
        </method>
        <method name="SetPosition">
          <arg name="TrackId" type="o" direction="in"/>
          <arg name="Position" type="x" direction="in"/>
        </method>
        <method name="OpenUri">
          <arg name="Uri" type="s" direction="in"/>
        </method>
        <signal name="Seeked">
          <arg name="Position" type="x"/>
        </signal>
        <property name="PlaybackStatus" type="s" access="read"/>
        <property name="LoopStatus" type="s" access="readwrite"/>
        <property name="Rate" type="d" access="readwrite"/>
        <property name="Shuffle" type="b" access="readwrite"/>
        <property name="Metadata" type="a{sv}" access="read"/>
        <property name="Volume" type="d" access="readwrite"/>
        <property name="Position" type="x" access="read"/>
        <property name="MinimumRate" type="d" access="read"/>
        <property name="MaximumRate" type="d" access="read"/>
        <property name="CanGoNext" type="b" access="read"/>
        <property name="CanGoPrevious" type="b" access="read"/>
        <property name="CanPlay" type="b" access="read"/>
        <property name="CanPause" type="b" access="read"/>
        <property name="CanSeek" type="b" access="read"/>
        <property name="CanControl" type="b" access="read"/>
      </interface>
    </node>
    """

    INTERFACE = "org.mpris.MediaPlayer2.Player"

    Seeked = signal()
    MinimumRate = MIN_RATE
    MaximumRate = MAX_RATE

    def _dbus_metadata(self) -> Optional[DbusMetadata]:
        metadata = self.adapter.metadata()
        
        if metadata:
            return get_dbus_metadata(metadata)

    def Next(self):
        logger.debug("%s.Next called", self.INTERFACE)
        if not self.CanGoNext:
            logger.debug("%s.Next not allowed", self.INTERFACE)
            return

        self.adapter.next()

    def Previous(self):
        logger.debug("%s.Previous called", self.INTERFACE)
        if not self.CanGoPrevious:
            logger.debug("%s.Previous not allowed", self.INTERFACE)
            return

        self.adapter.previous()

    def Pause(self):
        logger.debug("%s.Pause called", self.INTERFACE)
        if not self.CanPause:
            logger.debug("%s.Pause not allowed", self.INTERFACE)
            return

        self.adapter.pause()

    def PlayPause(self):
        logger.debug("%s.PlayPause called", self.INTERFACE)
        if not self.CanPause:
            logger.debug("%s.PlayPause not allowed", self.INTERFACE)
            return

        state = self.adapter.get_playstate()

        if state == PlayState.PLAYING:
            self.adapter.pause()
        elif state == PlayState.PAUSED:
            self.adapter.resume()
        elif state == PlayState.STOPPED:
            self.adapter.play()

    def Stop(self):
        logger.debug("%s.Stop called", self.INTERFACE)

        if not self.CanControl:
            logger.debug("%s.Stop not allowed", self.INTERFACE)
            return

        self.adapter.stop()

    def Play(self):
        logger.debug("%s.Play called", self.INTERFACE)

        if not self.CanPlay:
            logger.debug("%s.Play not allowed", self.INTERFACE)
            return

        state = self.adapter.get_playstate()

        if state == PlayState.PAUSED:
            self.adapter.resume()
        else:
            self.adapter.play()

    def Seek(self, offset: Microseconds):
        logger.debug("%s.Seek called", self.INTERFACE)

        if not self.CanSeek:
            logger.debug("%s.Seek not allowed", self.INTERFACE)
            return

        current_position = self.adapter.get_current_position()
        new_position = current_position + offset

        if new_position < BEGINNING:
            new_position = BEGINNING

        self.adapter.seek(new_position)

    def SetPosition(self, track_id: str, position: Microseconds):
        logger.debug("%s.SetPosition called", self.INTERFACE)

        if not self.CanSeek:
            logger.debug("%s.SetPosition not allowed", self.INTERFACE)
            return

        metadata = self.adapter.metadata()

        # use metadata from adapter if available
        if metadata and \
          'mpris:trackid' in metadata and \
          'mpris:length' in metadata:
            current_track = Track(
                track_id=metadata['mpris:trackid'],
                length=metadata['mpris:length']
            )

        # if no metadata, build metadata from Track interface
        else:
            current_track = self.adapter.get_current_track()

        if current_track is None:
            return

        if track_id != current_track.track_id:
            return

        if position < BEGINNING:
            return

        if current_track.length < position:
            return

        self.adapter.seek(position)

    def OpenUri(self, uri: str):
        logger.debug("%s.OpenUri called", self.INTERFACE)
        if not self.CanControl:
            logger.debug("%s.OpenUri not allowed", self.INTERFACE)
            return
        # NOTE Check if URI has MIME type known to the backend, if MIME support
        # is added to the backend.
        self.adapter.open_uri(uri)

    @property
    def PlaybackStatus(self):
        self.log_trace("Getting %s.PlaybackStatus", self.INTERFACE)
        state = self.adapter.get_playstate()
        return state.value.title()

    @property
    def LoopStatus(self):
        self.log_trace("Getting %s.LoopStatus", self.INTERFACE)
        if not self.adapter.is_repeating():
            return "None"

        else:
            if not self.adapter.is_playlist():
                return "Track"

            else:
                return "Playlist"

    @LoopStatus.setter
    def LoopStatus(self, value: str):
        if not self.CanControl:
            logger.debug("Setting %s.LoopStatus not allowed", self.INTERFACE)
            return
        logger.debug("Settideng %s.LoopStatus to %s", self.INTERFACE, value)

        self.adapter.set_loop_status(value)
        # if value == "None":
        # self.core.tracklist.set_repeat(False)
        # self.core.tracklist.set_single(False)
        # elif value == "Track":
        # self.core.tracklist.set_repeat(True)
        # self.core.tracklist.set_single(True)
        # elif value == "Playlist":
        # self.core.tracklist.set_repeat(True)
        # self.core.tracklist.set_single(False)

    @property
    def Rate(self):
        self.log_trace("Getting %s.Rate", self.INTERFACE)
        return self.adapter.get_rate()

    @Rate.setter
    def Rate(self, value: RateDecimal):
        if not self.CanControl:
            logger.debug("Setting %s.Rate not allowed", self.INTERFACE)
            return

        logger.debug("Setting %s.Rate to %s", self.INTERFACE, value)
        self.adapter.set_rate(value)

        if value == PAUSE_RATE:
            self.Pause()

    @property
    def Shuffle(self):
        self.log_trace("Getting %s.Shuffle", self.INTERFACE)
        return self.adapter.get_shuffle()

    @Shuffle.setter
    def Shuffle(self, value: bool):
        if not self.CanControl:
            logger.debug("Setting %s.Shuffle not allowed", self.INTERFACE)
            return

        logger.debug("Setting %s.Shuffle to %s", self.INTERFACE, value)
        self.adapter.set_shuffle(value)

    @property
    def Metadata(self):
        # prefer adapter's metadata to building our own
        metadata: DbusMetadata = self._dbus_metadata()

        if metadata:
            return metadata

        # build metadata if no metadata supplied by adapter
        self.log_trace("Building %s.Metadata", self.INTERFACE)
        track = self.adapter.get_current_track()
        stream_title = self.adapter.get_stream_title()

        if track is None:
            return DEFAULT_METADATA

        track_id = track.track_id
        res = {"mpris:trackid": Variant("o", track_id)}

        if track.length:
            res["mpris:length"] = Variant("x", track.length)

        if track.uri:
            res["xesam:url"] = Variant("s", track.uri)

        if stream_title or track.name:
            res["xesam:title"] = Variant("s", stream_title or track.name)

        if track.artists:
            artists = list(track.artists)
            artists.sort(key=lambda a: a.name or "")
            res["xesam:artist"] = Variant("as", [a.name for a in artists if a.name])

        if track.album and track.album.name:
            res["xesam:album"] = Variant("s", track.album.name)

        if track.album and track.album.artists:
            artists = list(track.album.artists)
            artists.sort(key=lambda a: a.name or "")
            res["xesam:albumArtist"] = Variant(
                "as", [a.name for a in artists if a.name]
            )

        art_url = self._get_art_url(track)
        if art_url:
            res["mpris:artUrl"] = Variant("s", art_url)
        if track.disc_no:
            res["xesam:discNumber"] = Variant("i", track.disc_no)
        if track.track_no:
            res["xesam:trackNumber"] = Variant("i", track.track_no)

        return res

    def _get_art_url(self, track: 'Track'):
        return self.adapter.get_art_url(track)

    @property
    def Volume(self):
        self.log_trace("Getting %s.Volume", self.INTERFACE)
        mute = self.adapter.is_mute()
        volume = self.adapter.get_volume()

        if volume is None or mute is True:
            return MUTE_VOL

        return volume

    @Volume.setter
    def Volume(self, value: VolumeDecimal):
        if not self.CanControl:
            logger.debug("Setting %s.Volume not allowed", self.INTERFACE)
            return
        logger.debug("Setting %s.Volume to %s", self.INTERFACE, value)

        if value is None:
            return

        if value < MUTE_VOL:
            value = MUTE_VOL
        elif value > MAX_VOL:
            value = MAX_VOL

        self.adapter.set_volume(value)

        if value > MUTE_VOL:
            self.adapter.set_mute(False)

    @property
    def Position(self):
        self.log_trace("Getting %s.Position", self.INTERFACE)
        return self.adapter.get_current_position()

    @property
    def CanGoNext(self):
        self.log_trace("Getting %s.CanGoNext", self.INTERFACE)
        if not self.CanControl:
            return False

        return self.adapter.can_go_next()

    @property
    def CanGoPrevious(self):
        self.log_trace("Getting %s.CanGoPrevious", self.INTERFACE)
        if not self.CanControl:
            return False

        return self.adapter.can_go_previous()

    @property
    def CanPlay(self):
        self.log_trace("Getting %s.CanPlay", self.INTERFACE)
        if not self.CanControl:
            return False

        return self.adapter.can_play()

    @property
    def CanPause(self):
        self.log_trace("Getting %s.CanPause", self.INTERFACE)
        if not self.CanControl:
            return False

        return True

    @property
    def CanSeek(self):
        self.log_trace("Getting %s.CanSeek", self.INTERFACE)
        if not self.CanControl:
            return False

        return True

    @property
    def CanControl(self):
        return self.adapter.can_control()



# Python and DBus compatibility
# See:  https://www.freedesktop.org/wiki/Specifications/mpris-spec/metadata/
import logging
from random import choices
from typing import Dict, Tuple, Any, Callable, Optional
from functools import wraps

from gi.repository.GLib import Variant
from unidecode import unidecode
from emoji import emoji_count, demojize

from .base import VALID_CHARS, Metadata, DbusMetadata, DbusTypes, \
    RAND_CHARS, NAME_PREFIX


logger = logging.getLogger(__name__)

# map of DBus metadata entries and their DBus types
METADATA_TYPES: Dict[str, str] = {
    "mpris:trackid": "o",
    "mpris:length": "x",
    "mpris:artUrl": "s",
    "xesam:url": "s",
    "xesam:title": "s",
    "xesam:artist": "as",
    "xesam:album": "s",
    "xesam:albumArtist": "as",
    "xesam:discNumber": 'i',
    "xesam:trackNumber": 'i',
    "xesam:comment": "as",
}

DBUS_NAME_MAX = 255
START_WITH = "_"
FIRST_CHAR = 0

# following must be subscriptable to be used with choices()
VALID_CHARS_SUB: Tuple[str] = tuple(VALID_CHARS)


def to_ascii(text: str) -> str:
  if emoji_count(text):
    text = demojize(text)

  return unidecode(text)


def random_name() -> str:
  rand = ''.join(choices(VALID_CHARS_SUB, k=RAND_CHARS))

  return NAME_PREFIX + rand


def enforce_dbus_length(func: Callable[..., str]) -> Callable[..., str]:
  @wraps(func)
  def new_func(*args, **kwargs) -> str:
    val = func(*args, **kwargs)

    return val[:DBUS_NAME_MAX]

  return new_func


@enforce_dbus_length
def get_dbus_name(
  name: Optional[str] = None,
  is_interface: bool = False
) -> str:
  if not name:
    return random_name()

  # interface names can contain hyphens
  if is_interface:
    valid_chars = {*VALID_CHARS, '-'}

  else:
    valid_chars = VALID_CHARS

  # convert utf8 to ascii
  new_name = to_ascii(name)

  # new name shouldn't have spaces
  new_name = new_name.replace(' ', '_')

  # new name should only contain DBus valid chars
  new_name = ''.join(char for char in new_name
                      if char in valid_chars)

  # DBus names can't start with numbers
  if new_name and new_name[FIRST_CHAR].isnumeric():
    # just stick an underscore in front of the number
    return START_WITH + new_name

  # but they can start with letters or underscore
  elif new_name:
    return new_name

  # if there is no name left after normalizing,
  # then make a random one and validate it
  return get_dbus_name(random_name())


def is_null_list(obj: Any) -> bool:
  if isinstance(obj, list):
    return all(item is None for item in obj)

  return False


def is_dbus_type(obj: Any) -> bool:
  return isinstance(obj, DbusTypes.__args__)


def is_valid_metadata(key: str, obj: Any) -> bool:
  if key not in METADATA_TYPES or obj is None:
    return False

  return is_dbus_type(obj) and not is_null_list(obj)


def get_dbus_metadata(metadata: Metadata) -> DbusMetadata:
  return {key: Variant(METADATA_TYPES[key], val)
          for key, val in metadata.items()
          if is_valid_metadata(key, val)}



import logging

from gi.repository import GLib
import pydbus

from .player import Player
from .playlists import Playlists
from .root import Root
from .base import NAME, BUS_TYPE
from .compat import get_dbus_name
from .adapters import MprisAdapter
from .tracklist import TrackList


logger = logging.getLogger(__name__)


class Server:
  def __init__(self,
               name: str = NAME,
               adapter: MprisAdapter = None):
    self.name = name
    self.adapter = adapter
    self.dbus_name = get_dbus_name(self.name)
    self.root = Root(name, self.adapter)
    self.player = Player(name, self.adapter)
    self.playlists = Playlists(name, self.adapter)
    self.tracklist = TrackList(name, self.adapter)

    self._loop = None
    self._publication_token = None

  def __del__(self):
    self.unpublish()

    if self._loop:
      self._loop.quit()

  def publish(self):
    bus_type = BUS_TYPE
    logger.debug(f"Connecting to D-Bus {bus_type} bus...")

    if bus_type == "system":
      bus = pydbus.SystemBus()

    else:
      bus = pydbus.SessionBus()

    logger.info(f"MPRIS server connected to D-Bus {bus_type} bus")

    self._publication_token = bus.publish(
      f"org.mpris.MediaPlayer2.{self.dbus_name}",
      ("/org/mpris/MediaPlayer2", self.root),
      ("/org/mpris/MediaPlayer2", self.player),
      ("/org/mpris/MediaPlayer2", self.playlists),
      ("/org/mpris/MediaPlayer2", self.tracklist)
    )

  def unpublish(self):
    if self._publication_token:
      self._publication_token.unpublish()
      self._publication_token = None

  def loop(self):
    if not self._publication_token:
      self.publish()

    self._loop = GLib.MainLoop()

    try:
      self._loop.run()

    finally:
      self._loop.quit()
      self._loop = None



from typing import Union, List, Tuple
import base64
import logging

from pydbus.generic import signal

from .base import PlaylistEntry, PlaylistValidity
from .interface import MprisInterface


logger = logging.getLogger(__name__)


class Playlists(MprisInterface):
    """
    <node>
      <interface name="org.mpris.MediaPlayer2.Playlists">
        <method name="ActivatePlaylist">
          <arg name="PlaylistId" type="o" direction="in"/>
        </method>
        <method name="GetPlaylists">
          <arg name="Index" type="u" direction="in"/>
          <arg name="MaxCount" type="u" direction="in"/>
          <arg name="Order" type="s" direction="in"/>
          <arg name="ReverseOrder" type="b" direction="in"/>
          <arg name="Playlists" type="a(oss)" direction="out"/>
        </method>
        <signal name="PlaylistChanged">
          <arg name="Playlist" type="oss"/>
        </signal>
        <property name="PlaylistCount" type="u" access="read"/>
        <property name="Orderings" type="as" access="read"/>
        <property name="ActivePlaylist" type="(b(oss))" access="read"/>
      </interface>
    </node>
    """

    INTERFACE = "org.mpris.MediaPlayer2.Playlists"
    PlaylistChanged = signal()

    def ActivatePlaylist(self, playlist_id: str):
        self.adapter.activate_playlist(playlist_id)
        # logger.debug(
        #     "%s.ActivatePlaylist(%r) called", self.INTERFACE, playlist_id
        # )
        # playlist_uri = get_playlist_uri(playlist_id)
        # playlist = self.core.playlists.lookup(playlist_uri).get()
        # if playlist and playlist.tracks:
        #     tl_tracks = self.core.tracklist.add(playlist.tracks).get()
        #     self.core.playback.play(tlid=tl_tracks[0].tlid).get()

    def GetPlaylists(self, index: int, max_count: int, order: str, reverse: bool) -> List[Tuple]:
        return self.adapter.get_playlists(index, max_count, order, reverse)
        # logger.debug(
        #     "%s.GetPlaylists(%r, %r, %r, %r) called",
        #     self.INTERFACE,
        #     index,
        #     max_count,
        #     order,
        #     reverse,
        # )
        # playlists = self.core.playlists.as_list().get()
        # if order == "Alphabetical":
        #     playlists.sort(key=lambda p: p.name, reverse=reverse)
        # elif order == "User" and reverse:
        #     playlists.reverse()
        # slice_end = index + max_count
        # playlists = playlists[index:slice_end]
        # results = [(get_playlist_id(p.uri), p.name, "") for p in playlists]
        # return results

    @property
    def PlaylistCount(self) -> int:
        self.log_trace("Getting %s.PlaylistCount", self.INTERFACE)
        return self.adapter.get_playlist_count()

    @property
    def Orderings(self) -> List[str]:
        self.log_trace("Getting %s.Orderings", self.INTERFACE)
        return self.adapter.get_orderings()

    @property
    def ActivePlaylist(self) -> Tuple[PlaylistValidity, PlaylistEntry]:
        return self.adapter.get_active_playlist()
        # self.log_trace("Getting %s.ActivePlaylist", self.INTERFACE)
        # playlist_is_valid = False
        # playlist = ("/", "None", "")
        # return (playlist_is_valid, playlist)

#
# def get_playlist_id(playlist_uri: Union[str, bytes]) -> str:
#     # Only A-Za-z0-9_ is allowed, which is 63 chars, so we can't use
#     # base64. Luckily, D-Bus does not limit the length of object paths.
#     # Since base32 pads trailing bytes with "=" chars, we need to replace
#     # them with an allowed character such as "_".
#     if isinstance(playlist_uri, str):
#         playlist_uri = playlist_uri.encode()
#     encoded_uri = base64.b32encode(playlist_uri).decode().replace("=", "_")
#     return "/com/mopidy/playlist/%s" % encoded_uri
#
#
# def get_playlist_uri(playlist_id: Union[str, bytes]) -> str:
#     if isinstance(playlist_id, bytes):
#         playlist_id = playlist_id.decode()
#     encoded_uri = playlist_id.split("/")[-1].replace("_", "=").encode()
#     return base64.b32decode(encoded_uri).decode()



from setuptools import setup
from pathlib import Path


REQS = \
  Path('requirements.txt') \
    .read_text() \
    .split('\n')

REQS = [
  req
  for req in REQS
  if not req.strip().startswith('#')
]

README = Path('README.md').read_text()


setup(
  name="mpris_server",
  version="0.2.17",
  description="Publish a MediaPlayer2 MPRIS device to D-Bus.",
  long_description=README,
  long_description_content_type="text/markdown",
  url="https://alexdelorenzo.dev",
  author="Alex DeLorenzo",
  license="AGPL-3.0",
  packages=['mpris_server'],
  zip_safe=True,
  install_requires=REQS,
  python_requires='>=3.6',
)



from typing import Optional
from abc import ABC
import logging

from pydbus.generic import signal

from .base import NAME, INTERFACE as MPRIS_INTERFACE


logger = logging.getLogger(__name__)

TRACE_LOG_LEVEL = 5


class MprisInterface(ABC):
    INTERFACE = MPRIS_INTERFACE

    PropertiesChanged = signal()

    def __init__(
      self,
      name: str = NAME,
      adapter: Optional['MprisAdapter'] = None
    ):
        self.name = name
        self.adapter = adapter

    def log_trace(self, *args, **kwargs):
        logger.log(TRACE_LOG_LEVEL, *args, **kwargs)



import logging

from .base import INTERFACE as _INTERFACE, NAME
from .interface import MprisInterface


logger = logging.getLogger(__name__)


class Root(MprisInterface):
    """
    <node>
      <interface name="org.mpris.MediaPlayer2">
        <method name="Raise"/>
        <method name="Quit"/>
        <property name="CanQuit" type="b" access="read"/>
        <property name="CanRaise" type="b" access="read"/>
        <property name="Fullscreen" type="b" access="readwrite"/>
        <property name="CanSetFullscreen" type="b" access="read"/>
        <property name="HasTrackList" type="b" access="read"/>
        <property name="Identity" type="s" access="read"/>
        <property name="DesktopEntry" type="s" access="read"/>
        <property name="SupportedUriSchemes" type="as" access="read"/>
        <property name="SupportedMimeTypes" type="as" access="read"/>
      </interface>
    </node>
    """

    INTERFACE = _INTERFACE

    def Raise(self):
        logger.debug("%s.Raise called", self.INTERFACE)
        self.adapter.set_raise(True)

    def Quit(self):
        logger.debug("%s.Quit called", self.INTERFACE)
        self.adapter.quit()

    @property
    def Fullscreen(self):
        self.log_trace("Getting %s.Fullscreen", self.INTERFACE)
        return self.adapter.get_fullscreen()

    @Fullscreen.setter
    def Fullscreen(self, value):
        logger.debug("Setting %s.Fullscreen to %s", self.INTERFACE, value)
        self.adapter.set_fullscreen(value)

    @property
    def DesktopEntry(self):
        self.log_trace("Getting %s.DesktopEntry", self.INTERFACE)
        return self.adapter.get_desktop_entry()

    @property
    def SupportedUriSchemes(self):
        self.log_trace("Getting %s.SupportedUriSchemes", self.INTERFACE)
        return self.adapter.get_uri_schemes()

    @property
    def SupportedMimeTypes(self):
        self.log_trace("Getting %s.SupportedMimeTypes", self.INTERFACE)
        return self.adapter.get_mime_types()

    @property
    def Identity(self) -> str:
        return self.name

    @property
    def CanQuit(self) -> bool:
        return self.adapter.can_quit()

    @property
    def CanRaise(self) -> bool:
        return self.adapter.can_raise()

    @property
    def CanSetFullscreen(self) -> bool:
        return self.adapter.can_fullscreen()

    @property
    def HasTrackList(self) -> bool:
        return self.adapter.has_tracklist()